
"""Contains class Population."""

import copy as cp
import pandas as pd
import numpy as np
import joblib as jl

from opqua.model import *

class Gillespie(object):
    """Class contains methods for simulating model with Gillespie algorithm.

    Class defines a model's events and methods for changing system state
    according to the possible events and simulating a timecourse using the
    Gillespie algorithm.

    Methods:
    getRates -- returns array containing rates for each event for a given system
        state
    doAction -- carries out an event, modifying system state
    run -- simulates model for a specified length of time
    """

    # Event ID constants:

    MIGRATE_HOST = 0
    MIGRATE_VECTOR = 1
    POPULATION_CONTACT_HOST_HOST = 2
    POPULATION_CONTACT_HOST_VECTOR = 3
    POPULATION_CONTACT_VECTOR_HOST = 4
    CONTACT_HOST_HOST = 5
    CONTACT_HOST_VECTOR = 6
    CONTACT_VECTOR_HOST = 7
    RECOVER_HOST = 8
    RECOVER_VECTOR = 9
    MUTATE_HOST = 10
    MUTATE_VECTOR = 11
    RECOMBINE_HOST = 12
    RECOMBINE_VECTOR = 13
    KILL_HOST = 14
    KILL_VECTOR = 15
    DIE_HOST = 16
    DIE_VECTOR = 17
    BIRTH_HOST = 18
    BIRTH_VECTOR = 19

    EVENT_IDS = { # must match above
        0:'MIGRATE_HOST',
        1:'MIGRATE_VECTOR',
        2:'POPULATION_CONTACT_HOST_HOST',
        3:'POPULATION_CONTACT_HOST_VECTOR',
        4:'POPULATION_CONTACT_VECTOR_HOST',
        5:'CONTACT_HOST_HOST',
        6:'CONTACT_HOST_VECTOR',
        7:'CONTACT_VECTOR_HOST',
        8:'RECOVER_HOST',
        9:'RECOVER_VECTOR',
        10:'MUTATE_HOST',
        11:'MUTATE_VECTOR',
        12:'RECOMBINE_HOST',
        13:'RECOMBINE_VECTOR',
        14:'KILL_HOST',
        15:'KILL_VECTOR',
        16:'DIE_HOST',
        17:'DIE_VECTOR',
        18:'BIRTH_HOST',
        19:'BIRTH_VECTOR'
        }

    def __init__(self, model):
        """Create a new Gillespie simulation object.

        Arguments:
        model -- the model this simulation belongs to (Model)
        """

        super(Gillespie, self).__init__() # initialize as parent class object

        # Event IDs
        self.evt_IDs = [
            self.MIGRATE_HOST, self.MIGRATE_VECTOR,
            self.POPULATION_CONTACT_HOST_HOST,
            self.POPULATION_CONTACT_HOST_VECTOR,
            self.POPULATION_CONTACT_VECTOR_HOST,
            self.CONTACT_HOST_HOST, self.CONTACT_HOST_VECTOR,
            self.CONTACT_VECTOR_HOST,
            self.RECOVER_HOST, self.RECOVER_VECTOR,
            self.MUTATE_HOST, self.MUTATE_VECTOR,
            self.RECOMBINE_HOST, self.RECOMBINE_VECTOR,
            self.KILL_HOST, self.KILL_VECTOR,
            self.DIE_HOST, self.DIE_VECTOR,
            self.BIRTH_HOST, self.BIRTH_VECTOR
            ]
            # event IDs in specific order to be used

        self.total_population_contact_rate_host = 0
        self.total_population_contact_rate_vector = 0

        self.model = model

    def getRates(self,population_ids):
        """Wrapper for calculating event rates as per current system state.

        Arguments:
        population_ids -- list with ids for every population in the model
            (list of Strings)

        Returns:
        Matrix with rates as values for events (rows) and populations (columns).
        Populations in order given in argument.
        """

        rates = np.zeros( [ len(self.evt_IDs), len(population_ids) ] )
            # rate array size of event space

        # Contact rates assume scaling area: large populations are equally
        # dense as small ones, so contact is constant with both host and
        # vector populations. If you don't want this to happen, modify each
        # population's base contact rate accordingly.

        # Now for each population...
        for i,id in enumerate(population_ids):
            # Calculate the rates:
            rates[self.MIGRATE_HOST,i] = (
                self.model.populations[id].total_migration_rate_hosts
                * self.model.populations[id].coefficients_hosts[
                    :, self.model.populations[id].MIGRATION
                    ].sum()
                )

            rates[self.MIGRATE_VECTOR,i] = (
                self.model.populations[id].total_migration_rate_vectors
                * self.model.populations[id].coefficients_vectors[
                    :, self.model.populations[id].MIGRATION
                    ].sum()
                )

            rates[self.POPULATION_CONTACT_HOST_HOST,i] = (
                np.sum([ list(
                    self.model.populations[id].neighbors_contact_hosts_hosts.values()
                    ) ])
                * np.multiply(
                    self.model.populations[id].coefficients_hosts[
                        :, self.model.populations[id].POPULATION_CONTACT
                        ],
                    self.model.populations[id].coefficients_hosts[
                        :, self.model.populations[id].INFECTED
                        ]
                    ).sum() / max( len( self.model.populations[id].hosts ), 1)
                * np.sum([
                    neighbor.neighbors_contact_hosts_hosts[self.model.populations[id]]
                    * neighbor.coefficients_hosts[
                        :, neighbor.RECEIVE_POPULATION_CONTACT
                        ].sum()
                    for neighbor in self.model.populations[id].neighbors_contact_hosts_hosts
                    ])
                )

            rates[self.POPULATION_CONTACT_HOST_VECTOR,i] = (
                np.sum([ list(
                    self.model.populations[id].neighbors_contact_hosts_vectors.values()
                    ) ])
                * np.multiply(
                    self.model.populations[id].coefficients_hosts[
                        :, self.model.populations[id].POPULATION_CONTACT
                        ],
                    self.model.populations[id].coefficients_hosts[
                        :, self.model.populations[id].INFECTED
                        ]
                    ).sum() / max( len( self.model.populations[id].hosts ), 1)
                * np.sum([
                    neighbor.neighbors_contact_vectors_hosts[self.model.populations[id]]
                    * neighbor.coefficients_vectors[
                        :, neighbor.RECEIVE_POPULATION_CONTACT
                        ].sum()
                    for neighbor in self.model.populations[id].neighbors_contact_hosts_vectors
                    ])
                )

            rates[self.POPULATION_CONTACT_VECTOR_HOST,i] = (
                np.sum([ list(
                    self.model.populations[id].neighbors_contact_vectors_hosts.values()
                    ) ])
                * np.multiply(
                    self.model.populations[id].coefficients_vectors[
                        :, self.model.populations[id].POPULATION_CONTACT
                        ],
                    self.model.populations[id].coefficients_vectors[
                        :, self.model.populations[id].INFECTED
                        ]
                    ).sum()
                * np.sum([
                    neighbor.neighbors_contact_hosts_vectors[self.model.populations[id]]
                    * neighbor.coefficients_hosts[
                        :, neighbor.RECEIVE_POPULATION_CONTACT
                        ].sum() / max( len( neighbor.hosts ), 1)
                    for neighbor in self.model.populations[id].neighbors_contact_vectors_hosts
                    ])
                )

            rates[self.CONTACT_HOST_HOST,i] = (
                self.model.populations[id].contact_rate_host_host
                * np.multiply(
                    self.model.populations[id].coefficients_hosts[
                        :, self.model.populations[id].CONTACT
                        ],
                    self.model.populations[id].coefficients_hosts[
                        :, self.model.populations[id].INFECTED
                        ]
                    ).sum()
                * self.model.populations[id].transmission_efficiency_host_host
                * np.sum( self.model.populations[id].coefficients_hosts[
                    :, self.model.populations[id].RECEIVE_CONTACT
                    ] )
                / max( len( self.model.populations[id].hosts ), 1)
                )

            rates[self.CONTACT_HOST_VECTOR,i] = (
                self.model.populations[id].contact_rate_host_vector
                * np.multiply(
                    self.model.populations[id].coefficients_hosts[
                        :, self.model.populations[id].CONTACT
                        ],
                    self.model.populations[id].coefficients_hosts[
                        :, self.model.populations[id].INFECTED
                        ]
                    ).sum()
                * self.model.populations[id].transmission_efficiency_host_vector
                * np.sum( self.model.populations[id].coefficients_vectors[
                    :, self.model.populations[id].RECEIVE_CONTACT
                    ] )
                / max( len( self.model.populations[id].hosts ), 1)
                )

            rates[self.CONTACT_VECTOR_HOST,i] = (
                self.model.populations[id].contact_rate_host_vector
                * np.multiply(
                    self.model.populations[id].coefficients_vectors[
                        :, self.model.populations[id].CONTACT
                        ],
                    self.model.populations[id].coefficients_vectors[
                        :, self.model.populations[id].INFECTED
                        ]
                    ).sum()
                * self.model.populations[id].transmission_efficiency_vector_host
                * np.sum( self.model.populations[id].coefficients_hosts[
                    :, self.model.populations[id].RECEIVE_CONTACT
                    ])
                / max( len( self.model.populations[id].hosts ), 1)
                )

            rates[self.RECOVER_HOST,i] = (
                self.model.populations[id].recovery_rate_host
                * self.model.populations[id].coefficients_hosts[
                    :, self.model.populations[id].RECOVERY
                    ].sum()
                )

            rates[self.RECOVER_VECTOR,i] = (
                self.model.populations[id].recovery_rate_vector
                * self.model.populations[id].coefficients_vectors[
                    :, self.model.populations[id].RECOVERY
                    ].sum()
                )

            rates[self.MUTATE_HOST,i] = (
                self.model.populations[id].mutate_in_host
                * self.model.populations[id].coefficients_hosts[
                    :, self.model.populations[id].MUTATION
                    ].sum()
                )

            rates[self.MUTATE_VECTOR,i] = (
                self.model.populations[id].mutate_in_vector
                * self.model.populations[id].coefficients_vectors[
                    :, self.model.populations[id].MUTATION
                    ].sum()
                )

            rates[self.RECOMBINE_HOST,i] = (
                self.model.populations[id].recombine_in_host
                * self.model.populations[id].coefficients_hosts[
                    :, self.model.populations[id].RECOMBINATION
                    ].sum()
                )

            rates[self.RECOMBINE_VECTOR,i] = (
                self.model.populations[id].recombine_in_vector
                * self.model.populations[id].coefficients_vectors[
                    :, self.model.populations[id].RECOMBINATION
                    ].sum()
                )

            rates[self.KILL_HOST,i] = (
                self.model.populations[id].lethality_rate_host
                * self.model.populations[id].coefficients_hosts[
                    :, self.model.populations[id].LETHALITY
                    ].sum()
                )

            rates[self.KILL_VECTOR,i] = (
                self.model.populations[id].lethality_rate_vector
                * self.model.populations[id].coefficients_vectors[
                    :, self.model.populations[id].LETHALITY
                    ].sum()
                )

            rates[self.DIE_HOST,i] = (
                self.model.populations[id].death_rate_host
                * len(self.model.populations[id].hosts)
                )

            rates[self.DIE_VECTOR,i] = (
                self.model.populations[id].death_rate_vector
                * len(self.model.populations[id].vectors)
                )

            rates[self.BIRTH_HOST,i] = (
                self.model.populations[id].birth_rate_host
                * self.model.populations[id].coefficients_hosts[
                    :, self.model.populations[id].NATALITY
                    ].sum()
                )

            rates[self.BIRTH_VECTOR,i] = (
                self.model.populations[id].birth_rate_vector
                * self.model.populations[id].coefficients_vectors[
                    :, self.model.populations[id].NATALITY
                    ].sum()
                )

        return rates

    def doAction(self,act,pop,rand):
        """Change system state according to act argument passed

        Arguments:
        act -- defines action to be taken, one of the event ID constants (int)
        pop -- population action will happen in (Population)
        rand -- random number used to define event (number 0-1)

        Returns:
        whether or not the model has changed state (Boolean)
        """

        changed = False

        if act == self.MIGRATE_HOST:
            rand = rand * pop.total_migration_rate_hosts
            r_cum = 0
            for neighbor in pop.neighbors_hosts:
                r_cum += pop.neighbors_hosts[neighbor]
                if r_cum > rand:
                    pop.migrate(neighbor,1,0, rand=(
                        ( rand - r_cum + pop.neighbors_hosts[neighbor] )
                        / pop.neighbors_hosts[neighbor] ) )
                    changed = True
                    break

        elif act == self.MIGRATE_VECTOR:
            rand = rand * pop.total_migration_rate_vectors
            r_cum = 0
            for neighbor in pop.neighbors_vectors:
                r_cum += pop.neighbors_vectors[neighbor]
                if r_cum > rand:
                    pop.migrate(neighbor,0,1, rand=(
                        ( rand - r_cum + pop.neighbors_vectors[neighbor] )
                        / pop.neighbors_vectors[neighbor] ) )
                    changed = True
                    break

        elif act == self.POPULATION_CONTACT_HOST_HOST:
            neighbors = list( pop.neighbors_contact_hosts_hosts.keys() )
            neighbor_rates = [
                pop.neighbors_contact_hosts_hosts[neighbor]
                * neighbor.neighbors_contact_hosts_hosts[pop]
                * neighbor.coefficients_hosts[
                    :, neighbor.RECEIVE_POPULATION_CONTACT
                    ].sum()
                / max( len( pop.hosts ), 1) # technically unnecessary
                for neighbor in neighbors
                ]
            rand = rand * np.sum(neighbor_rates)
            r_cum = 0
            for neighbor_i,neighbor_r in enumerate(neighbor_rates):
                r_cum += neighbor_r
                if r_cum > rand:
                    changed = pop.populationContact(
                        neighbors[neighbor_i],
                        ( rand - r_cum + neighbor_r ) / neighbor_r,
                        host_origin=True, host_target=True
                        )
                    break

        elif act == self.POPULATION_CONTACT_HOST_VECTOR:
            neighbors = list( pop.neighbors_contact_hosts_vectors.keys() )
            neighbor_rates = [
                pop.neighbors_contact_hosts_vectors[neighbor]
                * neighbor.neighbors_contact_vectors_hosts[pop]
                * neighbor.coefficients_vectors[
                    :, neighbor.RECEIVE_POPULATION_CONTACT
                    ].sum()
                / max( len( pop.hosts ), 1) # technically unnecessary
                for neighbor in neighbors
                ]
            rand = rand * np.sum(neighbor_rates)
            r_cum = 0
            for neighbor_i,neighbor_r in enumerate(neighbor_rates):
                r_cum += neighbor_r
                if r_cum > rand:
                    changed = pop.populationContact(
                        neighbors[neighbor_i],
                        ( rand - r_cum + neighbor_r ) / neighbor_r,
                        host_origin=True, host_target=False
                        )
                    break

        elif act == self.POPULATION_CONTACT_VECTOR_HOST:
            neighbors = list( pop.neighbors_contact_vectors_hosts.keys() )
            neighbor_rates = [
                pop.neighbors_contact_vectors_hosts[neighbor]
                * neighbor.neighbors_contact_hosts_vectors[pop]
                * neighbor.coefficients_hosts[
                    :, neighbor.RECEIVE_POPULATION_CONTACT
                    ].sum()
                / max( len( neighbor.hosts ), 1) # necessary!
                for neighbor in neighbors
                ]
            rand = rand * np.sum(neighbor_rates)
            r_cum = 0
            for neighbor_i,neighbor_r in enumerate(neighbor_rates):
                r_cum += neighbor_r
                if r_cum > rand:
                    changed = pop.populationContact(
                        neighbors[neighbor_i],
                        ( rand - r_cum + neighbor_r ) / neighbor_r,
                        host_origin=False, host_target=True
                        )
                    break

        elif act == self.CONTACT_HOST_HOST:
            changed = pop.contactHostHost(rand)

        elif act == self.CONTACT_HOST_VECTOR:
            changed = pop.contactHostVector(rand)

        elif act == self.CONTACT_VECTOR_HOST:
            changed = pop.contactVectorHost(rand)

        elif act == self.RECOVER_HOST:
            pop.recoverHost(rand)
            changed = True

        elif act == self.RECOVER_VECTOR:
            pop.recoverVector(rand)
            changed = True

        elif act == self.MUTATE_HOST:
            pop.mutateHost(rand)
            changed = True

        elif act == self.MUTATE_VECTOR:
            pop.mutateVector(rand)
            changed = True

        elif act == self.RECOMBINE_HOST:
            pop.recombineHost(rand)
            changed = True

        elif act == self.RECOMBINE_VECTOR:
            pop.recombineVector(rand)
            changed = True

        elif act == self.KILL_HOST:
            pop.killHost(rand)
            changed = True

        elif act == self.KILL_VECTOR:
            pop.killVector(rand)
            changed = True

        elif act == self.DIE_HOST:
            pop.dieHost(rand)
            changed = True

        elif act == self.DIE_VECTOR:
            pop.dieVector(rand)
            changed = True

        elif act == self.BIRTH_HOST:
            pop.birthHost(rand)
            changed = True

        elif act == self.BIRTH_VECTOR:
            pop.birthVector(rand)
            changed = True

        self.model.global_trackers['num_events'][self.EVENT_IDS[act]] += 1

        return changed

    def run(self,t0,tf,time_sampling=0,host_sampling=0,vector_sampling=0,
            print_every_n_events=1000):
        """Simulate model for a specified time between two time points.

        Simulates a time series using the Gillespie algorithm.

        Arguments:
        t0 -- initial time point to start simulation at (number)
        tf -- initial time point to end simulation at (number)
        time_sampling -- how many events to skip before saving a snapshot of the
            system state (saves all by default), if <0, saves only final state
            (int, default 0)
        host_sampling -- how many hosts to skip before saving one in a snapshot
            of the system state (saves all by default) (int, default 0)
        vector_sampling -- how many vectors to skip before saving one in a
            snapshot of the system state (saves all by default) (int, default 0)
        print_every_n_events -- number of events a message is printed to console
            (int>0, default 1000)

        Returns:
        dictionary containing model state history, with keys=times and
            values=Model objects with model snapshot at that time point
        """

        # Simulation variables
        self.model.t_var = t0 # keeps track of time
        history = { 0: self.model.copyState(
            host_sampling=host_sampling,
            vector_sampling=vector_sampling
            ) }
        intervention_tracker = 0
            # keeps track of what the next intervention should be
        self.model.interventions = sorted(
            self.model.interventions, key=lambda i: i.time
            )

        print_counter = 0 # only used to track when to print
        sampling_counter = 0 # used to track when to save a snapshot

        while self.model.t_var < tf: # repeat until t reaches end of timecourse
            population_ids = list( self.model.populations.keys() )
            r = self.getRates(population_ids) # get event rates in this state
            r_tot = np.sum(r) # sum of all rates

            # Time handling
            if r_tot > 0:
                dt = np.random.exponential( 1/r_tot ) # time until next event

                if (intervention_tracker < len(self.model.interventions)
                    and self.model.t_var + dt
                    >= self.model.interventions[intervention_tracker].time):
                    # if there are any interventions left and if it is time
                    # to make one,
                    while ( intervention_tracker < len(self.model.interventions)
                        and (self.model.t_var + dt
                        >= self.model.interventions[intervention_tracker].time
                        or r_tot == 0) and self.model.t_var < tf ):
                            # carry out all interventions at this time point,
                            # and additional timepoints if no events will happen
                        self.model.t_var = self.model.interventions[
                            intervention_tracker
                            ].time
                        self.model.interventions[
                            intervention_tracker
                            ].doIntervention()
                        intervention_tracker += 1 # advance the tracker

                        # save snapshot at this timepoint
                        sampling_counter = 0
                        history[self.model.t_var] = self.model.copyState()

                        # now recalculate rates
                        population_ids = list( self.model.populations.keys() )
                        r = self.getRates(population_ids)
                            # get event rates in this state
                        r_tot = np.sum(r) # sum of all rates

                    if r_tot > 0: # if no more events happening,
                        dt = np.random.exponential( 1/r_tot )
                            # time until next event
                    else:
                        self.model.t_var = tf # go to end

                self.model.t_var += dt # add time step to main timer

                # Event handling
                if self.model.t_var < tf: # if still within max time
                    u = np.random.random() * r_tot
                        # random uniform number between 0 and total rate
                    r_cum = 0 # cumulative rate
                    for e in range(r.shape[0]): # for every possible event,
                        for p in range(r.shape[1]):
                            # for every possible population,
                            r_cum += r[e,p]
                                # add this event's rate to cumulative rate
                            if u < r_cum:
                                # if random number is under cumulative rate

                                # print every n events
                                print_counter += 1
                                if print_counter == print_every_n_events:
                                    print_counter = 0
                                    print(
                                        'Simulating time: '
                                        + str(self.model.t_var) + ', event: '
                                        + self.EVENT_IDS[e]
                                        )

                                changed = self.doAction(
                                    e, self.model.populations[
                                        population_ids[p]
                                        ], ( u - r_cum + r[e,p] ) / r[e,p]
                                    ) # do corresponding action,
                                      # feed in renormalized random number

                                if changed: # if model state changed
                                    # update custom condition trackers
                                    for condition in self.model.custom_condition_trackers:
                                        if self.model.custom_condition_trackers[condition](self.model):
                                            self.model.global_trackers['custom_conditions'][condition].append(self.model.t_var)

                                    if time_sampling >= 0:
                                            # if state changed and saving history,
                                            # saves history at correct intervals
                                        sampling_counter += 1
                                        if sampling_counter > time_sampling:
                                            sampling_counter = 0
                                            history[self.model.t_var] = self.model.copyState(
                                                host_sampling=host_sampling,
                                                vector_sampling=vector_sampling
                                                )

                                break # exit event loop

                        else: # if the inner loop wasn't broken,
                            continue # continue outer loop

                        break # otherwise, break outer loop
            else: # if no events happening,
                if intervention_tracker < len(self.model.interventions):
                        # if still not done with interventions,
                    while (intervention_tracker < len(self.model.interventions)
                        and self.model.t_var
                        <= self.model.interventions[intervention_tracker].time
                        and self.model.t_var < tf):
                            # carry out all interventions at this time point
                        self.model.t_var = self.model.interventions[
                            intervention_tracker
                            ].time
                        self.model.interventions[
                            intervention_tracker
                            ].doIntervention()
                        intervention_tracker += 1 # advance the tracker
                else:
                    self.model.t_var = tf

        print( 'Simulating time: ' + str(self.model.t_var), 'END')
        history[tf] = self.model.copyState(
            host_sampling=host_sampling,
            vector_sampling=vector_sampling
            )
        history[tf].history = None
        history[tf].global_trackers = cp.deepcopy( self.model.global_trackers )

        return history




"""Contains class Population."""

import numpy as np
import copy as cp
import random

from opqua.internal.host import Host
from opqua.internal.vector import Vector

class Population(object):
    """Class defines a population with hosts, vectors, and specific parameters.

    Constants:
    # These all denote positions in coefficients_hosts and coefficients_vectors
    INFECTED -- position of "infected" Boolean values for each individual inside
        coefficients array
    CONTACT -- position of intra-population aggregated contact rate for each
        individual inside coefficients array
    RECEIVE_CONTACT -- position of intra-population aggregated receiving contact
        rate for each individual inside coefficients array
    LETHALITY -- position of aggregated death rate for each individual inside
        coefficients array
    NATALITY -- position of aggregated birth rate for each individual inside
        coefficients array
    RECOVERY -- position of aggregated recovery rate for each individual inside
        coefficients array
    MIGRATION -- position of aggregated inter-population migration rate for each
        individual inside coefficients array
    POPULATION_CONTACT -- position of inter-population aggregated contact rate
        for each individual inside coefficients array
    RECEIVE_POPULATION_CONTACT -- position of inter-population aggregated
        receiving contact rate for each individual inside coefficients array
    MUTATION -- position of aggregated mutation rate for each individual inside
        coefficients array
    RECOMBINATION -- position of aggregated recovery rate for each individual
        inside coefficients array
    NUM_COEFFICIENTS -- total number of types of coefficients (columns) in
        coefficient arrays
    CHROMOSOME_SEPARATOR -- character reserved to denote separate chromosomes in
        genomes

    Methods:
    copyState -- returns a slimmed-down version of the current population state
    setSetup -- assigns a given set of parameters to this population
    addHosts -- adds hosts to the population
    addVectors -- adds vectors to the population
    newHostGroup -- returns a list of random (healthy or any) hosts
    newVectorGroup -- returns a list of random (healthy or any) vectors
    removeHosts -- removes hosts from the population
    removeVectors -- removes vectors from the population
    addPathogensToHosts -- adds pathogens with specified genomes to hosts
    addPathogensToVectors -- adds pathogens with specified genomes to vectors
    treatHosts -- removes infections susceptible to given treatment from hosts
    treatVectors -- removes infections susceptible to treatment from vectors
    protectHosts -- adds protection sequence to hosts
    protectVectors -- adds protection sequence to vectors
    wipeProtectionHosts -- removes all protection sequences from hosts
    wipeProtectionVectors -- removes all protection sequences from hosts
    setHostMigrationNeighbor -- sets migration rate of hosts from this
        population towards another
    setVectorMigrationNeighbor -- sets migration rate of vectors from this
        population towards another
    migrate -- transfers hosts and/or vectors from this population to a neighbor
    setHostHostPopulationContactNeighbor -- set host-host contact rate from this
        population towards another one
    setHostVectorPopulationContactNeighbor -- set host-vector contact rate from
        this population towards another one
    setVectorHostPopulationContactNeighbor -- set vector-host contact rate from
        this population towards another one
    populationContact -- contacts hosts and/or vectors from this population to
        another
    contactHostHost -- contact any two (weighted) random hosts in population
    contactHostVector -- contact a (weighted) random host and vector in
        population
    contactVectorHost -- contact a (weighted) random vector and host in
        population
    recoverHost --removes all infections from given host
    recoverVector -- removes all infections from given vector
    killHost -- add host at this index to dead list, remove it from alive ones
    killVector -- add vector at this index to dead list, remove it from alive
        ones
    dieHost -- remove host at this index from alive lists
    dieVector -- remove vector at this index from alive lists
    birthHost -- add host at this index to population, remove it from alive ones
    birthVector -- add vector at this index to population, remove it from alive
        ones
    mutateHost -- mutates a single locus in a random pathogen in a host
    mutateVector -- mutates a single locus in a random pathogen in a vector
    recombineHost -- recombines two random pathogens in a host
    recombineVector -- recombines two random pathogens in a host
    updateHostCoefficients -- updates event coefficients in population's hosts
    updateVectorCoefficients -- updates event coefficients in population's
        vectors
    getWeightedRandom -- returns index of element chosen using weights from
        coefficient array and a given random number
    """

    INFECTED = 0
    CONTACT = 1
    RECEIVE_CONTACT = 2
    LETHALITY = 3
    NATALITY = 4
    RECOVERY = 5
    MIGRATION = 6
    POPULATION_CONTACT = 7
    RECEIVE_POPULATION_CONTACT = 8
    MUTATION = 9
    RECOMBINATION = 10

    NUM_COEFFICIENTS = 11

    CHROMOSOME_SEPARATOR = '/'

    def __init__(self, model, id, setup, num_hosts, num_vectors, slim=False):
        """Create a new Population.

        Arguments:
        model -- parent model this population is a part of (Model)
        id -- unique identifier for this population in the model (String)
        setup -- setup object with parameters for this population (Setup)
        num_hosts -- number of hosts to initialize population with (int)
        num_vectors -- number of hosts to initialize population with (int)

        Keyword arguments:
        slim -- whether to create a slimmed-down representation of the
            population for data storage (only ID, host and vector lists)
            (Boolean, default False)
        """
        super(Population, self).__init__()

        self.model = model
        self.id = id

        if not slim:
                # if not slimmed down for data storage, save other attributes

                # Each element in these following arrays contains the sum of all
                # pathogen rate coefficients within a host, weighted by fitness
                # and normalized to sum_fitness to obtain coefficient that
                # modifies the respective population-wide rate. Order given by
                # constants in Population class. Each host is one row.
            self.coefficients_hosts = np.zeros((1,self.NUM_COEFFICIENTS))
                # all weighted event rate coefficient modifiers for hosts
                # in population, first row is a dummy placeholder row
            self.coefficients_vectors = np.zeros((1,self.NUM_COEFFICIENTS))
                # all weighted event rate coefficient modifiers for vectors
                # in population, first row is a dummy placeholder row

            self.hosts = [
                Host(
                    self, self.id + '_' + str(id)
                ) for id in range(int(num_hosts))
                ]
                # contains all live hosts
            self.vectors = [
                Vector(
                    self, self.id + '_' + str(id)
                ) for id in range(int(num_vectors))
                ]
                # contains all live vectors

            # delete dummy first row, correct indexes of hosts afterwards
            # self.coefficients_hosts = np.delete(
            #     self.coefficients_hosts, 0, 0
            #     ) if num_hosts > 0 else self.coefficients_hosts
            # self.coefficients_vectors = np.delete(
            #     self.coefficients_vectors, 0, 0
            #     ) if num_vectors > 0 else self.coefficients_vectors
            # for h in self.hosts:
            #     h.coefficient_index -= 1
            # for v in self.vectors:
            #     v.coefficient_index -= 1

            self.infected_hosts = []
                # contains live, infected hosts (also in hosts list)
            self.healthy_hosts = self.hosts.copy()
                # contains live, healthy hosts (also in hosts list)
            self.infected_vectors = []
                # contains live, infected vectors (also in vectors list)
            self.dead_hosts = [] # contains dead hosts (not in hosts list)
            self.dead_vectors = [] # contains dead vectors (not in vectors list)

            self.neighbors_hosts = {} # dictionary with neighboring populations,
                # keys=Population, values=migration rate from this population to
                # neighboring population, for hosts only
            self.neighbors_vectors = {} # dictionary with neighbor populations,
                # keys=Population, values=migration rate from this population to
                # neighboring population, for vectors only
            self.total_migration_rate_hosts = 0 # sum of all migration rates
                # from this population to neighbors, hosts only
            self.total_migration_rate_vectors = 0 # sum of all migration rates
                # from this population to neighbors, vectors only

            self.neighbors_contact_hosts_hosts = {} # dictionary with
                # neighboring populations, keys=Population, values=population
                # contact rate from this population to neighboring population,
                # for hosts towards hosts only
            self.neighbors_contact_hosts_vectors = {} # dictionary with
                # neighboring populations, keys=Population, values=population
                # contact rate from this population to neighboring population,
                # for hosts towards vectors only
            self.neighbors_contact_vectors_hosts = {} # dictionary with neighbor
                # populations, keys=Population, values=population contact rate
                # from this population to neighboring population, for vectors
                # for vectors towards hosts only
            self.population_contact_rate_host = 0 # weighted sum of all host
                # population contact rates from this population to neighbors
            self.population_contact_rate_vector = 0 # weighted sum of all vector
                # population contact rates from this population to neighbors

            self.setSetup(setup)

            self.setHostMigrationNeighbor(self,0)
            self.setVectorMigrationNeighbor(self,0)
            self.setHostHostPopulationContactNeighbor(self,0)
            self.setHostVectorPopulationContactNeighbor(self,0)
            self.setVectorHostPopulationContactNeighbor(self,0)

    def copyState(self,host_sampling=0,vector_sampling=0):
        """Returns a slimmed-down version of the current population state.

        Arguments:
        host_sampling -- how many hosts to skip before saving one in a snapshot
            of the system state (saves all by default) (int, default 0)
        vector_sampling -- how many vectors to skip before saving one in a
            snapshot of the system state (saves all by default) (int, default 0)

        Returns:
        Population object with current host and vector lists.
        """

        copy = Population(self.model, self.id, None, 0, 0, slim=True)
        if host_sampling > 0:
            host_sample = random.sample(
                self.hosts, int( len(self.hosts) / (host_sampling+1) )
                )
            dead_host_sample = random.sample(
                self.dead_hosts, int( len(self.dead_hosts) / (host_sampling+1) )
                )
            copy.hosts = [ h.copyState() for h in host_sample ]
            copy.dead_hosts = [ h.copyState() for h in dead_host_sample ]
        else:
            copy.hosts = [ h.copyState() for h in self.hosts ]
            copy.dead_hosts = [ h.copyState() for h in self.dead_hosts ]

        if vector_sampling > 0:
            vector_sample = random.sample(
                self.vectors, int( len(self.vectors) / (vector_sampling+1) )
                )
            dead_vector_sample = random.sample(
                self.dead_vectors,int(len(self.dead_vectors)/(vector_sampling+1))
                )
            copy.vectors = [ v.copyState() for v in vector_sample ]
            copy.dead_vectors = [ v.copyState() for v in dead_vector_sample ]
        else:
            copy.vectors = [ v.copyState() for v in self.vectors ]
            copy.dead_vectors = [ v.copyState() for v in self.dead_vectors ]

        return copy

    def setSetup(self, setup):
        """Assign parameters stored in Setup object to this population.

        Arguments:
        setup -- the setup to be assigned (Setup)
        """

        self.setup = setup

        self.num_loci = setup.num_loci
        self.possible_alleles = setup.possible_alleles

        self.fitnessHost = setup.fitnessHost
        self.contactHost = setup.contactHost
        self.receiveContactHost = setup.receiveContactHost
        self.lethalityHost = setup.lethalityHost
        self.natalityHost = setup.natalityHost
        self.recoveryHost = setup.recoveryHost
        self.migrationHost = setup.migrationHost
        self.populationContactHost = setup.populationContactHost
        self.receivePopulationContactHost = setup.receivePopulationContactHost
        self.mutationHost = setup.mutationHost
        self.recombinationHost = setup.recombinationHost
        self.updateHostCoefficients()

        self.fitnessVector = setup.fitnessVector
        self.contactVector = setup.contactVector
        self.receiveContactVector = setup.receiveContactVector
        self.lethalityVector = setup.lethalityVector
        self.natalityVector = setup.natalityVector
        self.recoveryVector = setup.recoveryVector
        self.migrationVector = setup.migrationVector
        self.populationContactVector = setup.populationContactVector
        self.receivePopulationContactVector = setup.receivePopulationContactVector
        self.mutationVector = setup.mutationVector
        self.recombinationVector = setup.recombinationVector
        self.updateVectorCoefficients()

        self.contact_rate_host_vector = setup.contact_rate_host_vector
        self.contact_rate_host_host = setup.contact_rate_host_host
            # contact rate assumes fixed area--large populations are dense
            # populations, so contact scales linearly with both host and vector
            # populations. If you don't want this to happen, modify the
            # population's contact rate accordingly.
        self.transmission_efficiency_host_vector = setup.transmission_efficiency_host_vector
        self.transmission_efficiency_vector_host = setup.transmission_efficiency_vector_host
        self.transmission_efficiency_host_host = setup.transmission_efficiency_host_host
        self.mean_inoculum_host = setup.mean_inoculum_host
        self.mean_inoculum_vector = setup.mean_inoculum_vector
        self.recovery_rate_host = setup.recovery_rate_host
        self.recovery_rate_vector = setup.recovery_rate_vector
        self.lethality_rate_host = setup.lethality_rate_host
        self.lethality_rate_vector = setup.lethality_rate_vector
        self.recombine_in_host = setup.recombine_in_host
        self.recombine_in_vector = setup.recombine_in_vector
        self.num_crossover_host = setup.num_crossover_host
        self.num_crossover_vector = setup.num_crossover_vector
        self.mutate_in_host = setup.mutate_in_host
        self.mutate_in_vector = setup.mutate_in_vector
        self.death_rate_host = setup.death_rate_host
        self.death_rate_vector = setup.death_rate_vector
        self.birth_rate_host = setup.birth_rate_host
        self.birth_rate_vector = setup.birth_rate_vector
        self.vertical_transmission_host = setup.vertical_transmission_host
        self.vertical_transmission_vector = setup.vertical_transmission_vector
        self.inherit_protection_host = setup.inherit_protection_host
        self.inherit_protection_vector = setup.inherit_protection_vector
        self.protection_upon_recovery_host \
            = setup.protection_upon_recovery_host
        self.protection_upon_recovery_vector \
            = setup.protection_upon_recovery_vector

    def addHosts(self, num_hosts):
        """Add a number of healthy hosts to population, return list with them.

        Arguments:
        num_hosts -- number of hosts to be added (int)

        Returns:
        list containing new hosts
        """

        new_hosts = [
            Host(
                self, self.id + '_' + str( i + len(self.hosts) )
                ) for i in range(num_hosts)
            ]
        self.hosts += new_hosts
        self.healthy_hosts += new_hosts

        return new_hosts

    def addVectors(self, num_vectors):
        """Add a number of healthy vectors to population, return list with them.

        Arguments:
        num_vectors -- number of vectors to be added (int)

        Returns:
        list containing new vectors
        """

        new_vectors = [
            Vector(
                self, self.id + '_' + str( i + len(self.vectors) )
                ) for i in range(num_vectors)
            ]
        self.vectors += new_vectors

        return new_vectors

    def newHostGroup(self, hosts=-1, type='any'):
        """Return a list of random hosts in population.

        Arguments:
        hosts -- number of hosts to be sampled randomly: if <0, samples from
            whole population; if <1, takes that fraction of population; if >=1,
            samples that integer number of hosts (default -1, number)

        Keyword arguments:
        type -- whether to sample healthy hosts only, infected hosts only, or
            any hosts (default 'any'; String = {'healthy', 'infected', 'any'})

        Returns:
        list containing sampled hosts
        """

        possible_hosts = []

        if type=='healthy':
            possible_hosts = self.healthy_hosts
        elif type=='infected':
            possible_hosts = self.infected_hosts
        elif type=='any':
            possible_hosts = self.hosts
        else:
            raise ValueError(
                '"' + str(type)
                + '" is not a type of host for newHostGroup.'
                )

        num_hosts = -1
        if hosts < 0:
            num_hosts = len(possible_hosts)
        elif hosts < 1:
            num_hosts = int( hosts * len(possible_hosts) )
        else:
            num_hosts = hosts

        if len(possible_hosts) >= num_hosts:
            return np.random.choice(possible_hosts, num_hosts, replace=False)
        else:
            raise ValueError(
                "You're asking for " + str(num_hosts)
                + '"' + type + '" hosts, but population ' + str(self.id)
                + " only has " + str( len(self.healthy_hosts) ) + "."
                )

    def newVectorGroup(self, vectors=-1, type='any'):
        """Return a list of random vectors in population.

        Arguments:
        vectors -- number of vectors to be sampled randomly: if <0, samples from
            whole population; if <1, takes that fraction of population; if >=1,
            samples that integer number of vectors (default -1, number)

        Keyword arguments:
        type -- whether to sample healthy vectors only, infected vectors
            only, or any vectors (default 'any'; String = {'healthy',
            'infected', 'any'})

        Returns:
        list containing sampled vectors
        """

        possible_vectors = []
        if type=='healthy':
            for vector in self.vectors:
                if vector not in self.infected_vectors:
                    possible_vectors.append(vector)
        elif type=='infected':
            possible_vectors = self.infected_vectors
        elif type=='any':
            possible_vectors = self.vectors
        else:
            raise ValueError(
                '"' + str(type)
                + '" is not a type of vector for newVectorGroup.'
                )

        num_vectors = -1
        if vectors < 0:
            num_vectors = len(possible_vectors)
        elif vectors < 1:
            num_vectors = int( vectors * len(possible_vectors) )
        else:
            num_vectors = vectors

        if len(possible_vectors) >= num_vectors:
            return np.random.choice(possible_vectors, num_vectors, replace=False)
        else:
            raise ValueError(
                "You're asking for " + str(num_vectors)
                + '"' + type + '" vectors, but population ' + str(self.id)
                + " only has "
                + str( len(self.vectors) - len(self.infected_vectors) )
                + "."
                )

    def removeHosts(self, num_hosts_or_list):
        """Remove a number of specified or random hosts from population.

        Arguments:
        num_hosts_or_list -- number of hosts to be sampled randomly for removal
            or list of hosts to be removed, must be hosts in this population
            (int or list of Hosts)
        """

        if isinstance(num_hosts_or_list, list):
            for host_removed in num_hosts_or_list:
                if host_removed in self.hosts:
                    if host_removed in self.infected_hosts:
                        self.infected_hosts.remove( host_removed )
                    else:
                        self.healthy_hosts.remove( host_removed )

                    self.hosts.remove( host_removed )
                    for h in self.hosts:
                        if h.coefficient_index > host_removed.coefficient_index:
                            h.coefficient_index -= 1

                    self.coefficients_hosts = np.delete(
                        self.coefficients_hosts,
                        host_removed.coefficient_index, 0
                        )
        else:
            for _ in range(num_hosts_or_list):
                host_removed = np.random.choice(self.hosts)
                if host_removed in self.infected_hosts:
                    self.infected_hosts.remove( host_removed )
                else:
                    self.healthy_hosts.remove( host_removed )

                self.hosts.remove( host_removed )
                for h in self.hosts:
                    if h.coefficient_index > host_removed.coefficient_index:
                        h.coefficient_index -= 1

                self.coefficients_hosts = np.delete(
                    self.coefficients_hosts,
                    host_removed.coefficient_index, 0
                    )

    def removeVectors(self, num_vectors_or_list):
        """Remove a number of specified or random vectors from population.

        Arguments:
        num_vectors_or_list -- number of vectors to be sampled randomly for
            removal or list of vectors to be removed, must be vectors in this
            population (int or list of Vectors)
        """

        if isinstance(num_vectors_or_list, list):
            for vector_removed in num_vectors_or_list:
                if vector_removed in self.vectors:
                    if vector_removed in self.infected_vectors:
                        self.infected_vectors.remove( vector_removed )

                    self.vectors.remove( vector_removed )
                    for v in self.vectors:
                        if v.coefficient_index>vector_removed.coefficient_index:
                            v.coefficient_index -= 1

                    self.coefficients_vectors = np.delete(
                        self.coefficients_vectors,
                        vector_removed.coefficient_index, 0
                        )
        else:
            for _ in range(num_vectors):
                vector_removed = np.random.choice(self.vectors)
                if vector_removed in self.infected_vectors:
                    self.infected_vectors.remove( vector_removed )

                self.vectors.remove( vector_removed )
                for v in self.vectors:
                    if v.coefficient_index > vector_removed.coefficient_index:
                        v.coefficient_index -= 1

                self.coefficients_vectors = np.delete(
                    self.coefficients_vectors,
                    vector_removed.coefficient_index, 0
                    )

    def addPathogensToHosts(self, genomes_numbers, hosts=[]):
        """Add specified pathogens to random hosts, optionally from a list.

        Arguments:
        genomes_numbers -- dictionary conatining pathogen genomes to add as keys
            and number of hosts each one will be added to as values (dict with
            keys=Strings, values=int)

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        if len(hosts) == 0:
            hosts = self.hosts

        for genome in genomes_numbers:
            if len(genome) == self.num_loci and genomes_numbers[genome]>0 and all( [
                allele in self.possible_alleles[i] + self.CHROMOSOME_SEPARATOR
                    for i,allele in enumerate(genome)
                ] ):
                rand_hosts = np.random.choice(
                    hosts, int(genomes_numbers[genome]), replace=False
                    )

                for host in rand_hosts:
                    host.acquirePathogen(genome)

                if genome not in self.model.global_trackers['genomes_seen']:
                    self.model.global_trackers['genomes_seen'].append(genome)

            elif genomes_numbers[genome]>0:
                raise ValueError('Genome ' + genome + ' must be of length '
                    + str(self.num_loci)
                    + ' and contain only the following characters at each '
                    + 'position: ' + str(self.possible_alleles)
                    + ', or the chromosome separator character "'
                    + self.CHROMOSOME_SEPARATOR + '" .')

    def addPathogensToVectors(self, genomes_numbers, vectors=[]):
        """Add specified pathogens to random vectors, optionally from a list.

        Arguments:
        genomes_numbers -- dictionary conatining pathogen genomes to add as keys
            and number of vectors each one will be added to as values (dict with
            keys=Strings, values=int)

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples
            from whole population (default empty list; empty)
        """

        if len(vectors) == 0:
            vectors = self.vectors

        for genome in genomes_numbers:
            if len(genome) == self.num_loci and genomes_numbers[genome]>0 and all( [
                allele in self.possible_alleles[i] + self.CHROMOSOME_SEPARATOR
                    for i,allele in enumerate(genome)
                ] ):
                rand_vectors = np.random.choice(
                    vectors, int(genomes_numbers[genome]), replace=False
                    )

                for vector in rand_vectors:
                    vector.acquirePathogen(genome)

                if genome not in self.model.global_trackers['genomes_seen']:
                    self.model.global_trackers['genomes_seen'].append(genome)

            elif genomes_numbers[genome]>0:
                raise ValueError('Genome ' + genome + ' must be of length '
                    + str(self.num_loci)
                    + ' and contain only the following characters at each '
                    + 'position: ' + self.possible_alleles
                    + self.CHROMOSOME_SEPARATOR + ' .')

    def treatHosts(self, frac_hosts, resistance_seqs, hosts=[]):
        """Treat random fraction of infected hosts against some infection.

        Removes all infections with genotypes susceptible to given treatment.
        Pathogens are removed if they are missing at least one of the sequences
        in resistance_seqs from their genome. Removes this organism from
        population infected list and adds to healthy list if appropriate.

        Arguments:
        frac_hosts -- fraction of hosts considered to be randomly selected
            (number between 0 and 1)
        resistance_seqs -- contains sequences required for treatment resistance
            (list of Strings)

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        hosts_to_consider = self.hosts
        if len(hosts) > 0:
            hosts_to_consider = hosts

        possible_infected_hosts = []
        for host in hosts_to_consider:
            if len( host.pathogens ):
                possible_infected_hosts.append( host )

        treat_hosts = np.random.choice(
            possible_infected_hosts,
            int( frac_hosts * len( possible_infected_hosts ) ), replace=False
            )
        for host in treat_hosts:
            host.applyTreatment(resistance_seqs)

    def treatVectors(self, frac_vectors, resistance_seqs, vectors=[]):
        """Treat random fraction of infected vectors agains some infection.

        Removes all infections with genotypes susceptible to given treatment.
        Pathogens are removed if they are missing at least one of the sequences
        in resistance_seqs from their genome. Removes this organism from
        population infected list and adds to healthy list if appropriate.

        Arguments:
        frac_vectors -- fraction of vectors considered to be randomly selected
            (number between 0 and 1)
        resistance_seqs -- contains sequences required for treatment resistance
            (list of Strings)

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples
            from whole population (default empty list; empty)
        """

        vectors_to_consider = self.vectors
        if len(vectors) > 0:
            vectors_to_consider = vectors

        possible_infected_vectors = []
        for vector in vectors_to_consider:
            if len( vector.pathogens ):
                possible_infected_vectors.append( vector )

        treat_vectors = np.random.choice(
            possible_infected_vectors,
            int( frac_vectors * len( possible_infected_vectors ) ),
            replace=False
            )
        for vector in treat_vectors:
            vector.applyTreatment(resistance_seqs)

    def protectHosts(self, frac_hosts, protection_sequence, hosts=[]):
        """Protect a random fraction of infected hosts against some infection.

        Adds protection sequence specified to a random fraction of the hosts
        specified. Does not cure them if they are already infected.

        Arguments:
        frac_hosts -- fraction of hosts considered to be randomly selected
            (number between 0 and 1)
        protection_sequence -- sequence against which to protect (String)

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        hosts_to_consider = self.hosts
        if len(hosts) > 0:
            hosts_to_consider = hosts

        protect_hosts = np.random.choice(
            self.hosts, int( frac_hosts * len( hosts_to_consider ) ),
            replace=False
            )
        for host in protect_hosts:
            host.protection_sequences.append(protection_sequence)

    def protectVectors(self, frac_vectors, protection_sequence, vectors=[]):
        """Protect a random fraction of infected vectors against some infection.

        Adds protection sequence specified to a random fraction of the vectors
        specified. Does not cure them if they are already infected.

        Arguments:
        frac_vectors -- fraction of vectors considered to be randomly selected
            (number between 0 and 1)
        protection_sequence -- sequence against which to protect (String)

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples
            from whole population (default empty list; empty)
        """

        vectors_to_consider = self.vectors
        if len(vectors) > 0:
            vectors_to_consider = vectors

        protect_vectors = np.random.choice(
            self.vectors, int( frac_vectors * len( vectors_to_consider ) ),
            replace=False
            )
        for vector in protect_vectors:
            vector.protection_sequences.append(protection_sequence)

    def wipeProtectionHosts(self, hosts=[]):
        """Removes all protection sequences from hosts.

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        hosts_to_consider = self.hosts
        if len(hosts) > 0:
            hosts_to_consider = hosts

        for host in hosts_to_consider:
            host.protection_sequences = []

    def wipeProtectionVectors(self, vectors=[]):
        """Removes all protection sequences from vectors.

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        vectors_to_consider = self.vectors
        if len(vectors) > 0:
            vectors_to_consider = vectors

        for vector in vectors_to_consider:
            vector.protection_sequences = []

    def setHostMigrationNeighbor(self, neighbor, rate):
        """Set host migration rate from this population towards another one.

         Arguments:
         neighbor -- population towards which migration rate will be specified
            (Population)
         rate -- migration rate from this population to the neighbor (number)
         """

        if neighbor in self.neighbors_hosts:
            self.total_migration_rate_hosts -= self.neighbors_hosts[neighbor]

        self.neighbors_hosts[neighbor] = rate
        self.total_migration_rate_hosts += rate

    def setVectorMigrationNeighbor(self, neighbor, rate):
        """Set vector migration rate from this population towards another one.

         Arguments:
         neighbor -- population towards which migration rate will be specified
            (Population)
         rate -- migration rate from this population to the neighbor (number)
         """

        if neighbor in self.neighbors_vectors:
            self.total_migration_rate_vectors -= self.neighbors_vectors[neighbor]

        self.neighbors_vectors[neighbor] = rate
        self.total_migration_rate_vectors += rate

    def migrate(self, target_pop, num_hosts, num_vectors, rand=None):
        """Transfer hosts and/or vectors from this population to another.

        Arguments:
        target_pop -- population towards which migration will occur (Population)
        num_hosts -- number of hosts to transfer (int)
        num_vectors -- number of vectors to transfer (int)

        Keyword arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individuals to migrate; if None, generates new random number to
            choose (through numpy), otherwise, assumes event is happening
            through Gillespie class call and migrates a single host or vector
            (default None; 0-1 number)
        """

        if rand is None:
            migrating_hosts = np.random.choice(
                self.hosts, num_hosts, replace=False,
                p=self.coefficients_hosts[:,self.MIGRATION]
                )
            migrating_vectors = np.random.choice(
                self.vectors, num_vectors, replace=False,
                p=self.coefficients_vectors[:,self.MIGRATION]
                )
        elif num_hosts == 1:
            index_host,rand = self.getWeightedRandom(
                rand, self.coefficients_hosts[:,self.MIGRATION]
                )
            migrating_hosts = [self.hosts[index_host]]
            migrating_vectors = []
        else:
            index_vector,rand = self.getWeightedRandom(
                rand, self.coefficients_vectors[:,self.MIGRATION]
                )
            migrating_hosts = []
            migrating_vectors = [self.vectors[index_vector]]


        for host in migrating_hosts:
            genomes = { g:1 for g in host.pathogens }
            self.removeHosts([host])
            new_host_list = target_pop.addHosts(1) # list of 1
            target_pop.addPathogensToHosts(genomes,hosts=new_host_list)
            host = None

        for vector in migrating_vectors:
            genomes = { g:1 for g in vector.pathogens }
            self.removeVectors([vector])
            new_vector_list = target_pop.addVectors(1) # list of 1
            target_pop.addPathogensToVectors(genomes,vectors=new_vector_list)
            vector = None

    def setHostHostPopulationContactNeighbor(self, neighbor, rate):
        """Set host-host contact rate from this population towards another one.

         Arguments:
         neighbor -- population towards which migration rate will be specified
            (Population)
         rate -- migration rate from this population to the neighbor (number)
         """

        self.neighbors_contact_hosts_hosts[neighbor] = rate

    def setHostVectorPopulationContactNeighbor(self, neighbor, rate):
        """Set host-vector contact rate from this population to another one.

         Arguments:
         neighbor -- population towards which migration rate will be specified
            (Population)
         rate -- migration rate from this population to the neighbor (number)
         """

        self.neighbors_contact_hosts_vectors[neighbor] = rate

    def setVectorHostPopulationContactNeighbor(self, neighbor, rate):
        """Set vector-host contact rate from this population to another one.

         Arguments:
         neighbor -- population towards which migration rate will be specified
            (Population)
         rate -- migration rate from this population to the neighbor (number)
         """

        self.neighbors_contact_vectors_hosts[neighbor] = rate

    def populationContact(
        self, target_pop, rand, host_origin=True, host_target=True):
        """Contacts hosts and/or vectors from this population to another.

        Arguments:
        target_pop -- population towards which migration will occur (Population)
        rand -- uniform random number from 0 to 1 to use when choosing
            individuals to contact

        Keyword arguments:
        host_origin -- whether to draw from hosts in the origin population
            (as opposed to vectors) (Boolean)
        host_target -- whether to draw from hosts in the target population
            (as opposed to vectors) (Boolean)
        """

        if host_origin:
            index_host,rand = self.getWeightedRandom(
                rand, np.multiply(
                    self.coefficients_hosts[:,self.POPULATION_CONTACT],
                    self.coefficients_hosts[:,self.INFECTED]
                    )
                )
            origin = self.hosts[index_host]
        else:
            index_vector,rand = self.getWeightedRandom(
                rand, np.multiply(
                    self.coefficients_vectors[:,self.POPULATION_CONTACT],
                    self.coefficients_vectors[:,self.INFECTED]
                    )
                )
            origin = self.vectors[index_vector]

        if host_target:
            index_host,rand = target_pop.getWeightedRandom(
                rand,
                target_pop.coefficients_hosts[
                    :,target_pop.RECEIVE_POPULATION_CONTACT
                    ]
                )
            origin.infectHost(target_pop.hosts[index_host])
        else:
            index_vector,rand = target_pop.getWeightedRandom(
                rand,
                target_pop.coefficients_vectors[
                    :,target_pop.RECEIVE_POPULATION_CONTACT
                    ]
                )
            origin.infectVector(target_pop.vectors[index_vector])

    def contactHostHost(self, rand):
        """Contact any two (weighted) random hosts in population.

        Carries out possible infection events from the first organism into the
        second.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individuals to contact

        Returns:
        whether or not the model has changed state (Boolean)
        """

        index_host,rand = self.getWeightedRandom(
            rand, np.multiply(
                self.coefficients_hosts[:,self.CONTACT],
                self.coefficients_hosts[:,self.INFECTED]
                )
            )
        index_other_host,rand = self.getWeightedRandom(
            rand, self.coefficients_hosts[:,self.RECEIVE_CONTACT]
            )

        changed = self.hosts[index_host].infectHost(
            self.hosts[index_other_host]
            )

        return changed

    def contactHostVector(self, rand):
        """Contact a (weighted) random host and vector in population.

        Carries out possible infection events from the first organism into the
        second.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individuals to contact

        Returns:
        whether or not the model has changed state (Boolean)
        """

        index_host,rand = self.getWeightedRandom(
            rand, np.multiply(
                self.coefficients_hosts[:,self.CONTACT],
                self.coefficients_hosts[:,self.INFECTED]
                )
            )
        index_vector,rand = self.getWeightedRandom(
            rand, self.coefficients_vectors[:,self.RECEIVE_CONTACT]
            )
        changed = self.hosts[index_host].infectVector(
            self.vectors[index_vector]
            )

        return changed

    def contactVectorHost(self, rand):
        """Contact a (weighted) random vector and host in population.

        Carries out possible infection events from the first organism into the
        second.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individuals to contact

        Returns:
        whether or not the model has changed state (Boolean)
        """

        index_vector,rand = self.getWeightedRandom(
            rand, np.multiply(
                self.coefficients_vectors[:,self.CONTACT],
                self.coefficients_vectors[:,self.INFECTED]
                )
            )
        index_host,rand = self.getWeightedRandom(
            rand,self.coefficients_hosts[:,self.RECEIVE_CONTACT]
            )

        changed = self.vectors[index_vector].infectHost(
            self.hosts[index_host]
            )

        return changed

    def recoverHost(self, rand):
        """Remove all infections from host at this index.

        If model is protecting upon recovery, add protecion sequence as defined
        by the indexes in the corresponding model parameter. Remove from
        population infected list and add to healthy list.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual to recover
        """

        index_host,rand = self.getWeightedRandom(
            rand,self.coefficients_hosts[:,self.RECOVERY]
            )

        self.hosts[index_host].recover()

    def recoverVector(self, rand):
        """Remove all infections from vector at this index.

        If model is protecting upon recovery, add protecion sequence as defined
        by the indexes in the corresponding model parameter. Remove from
        population infected list and add to healthy list.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual to recover
        """

        index_vector,rand = self.getWeightedRandom(
            rand,self.coefficients_vectors[:,self.RECOVERY]
            )

        self.vectors[index_vector].recover()

    def killHost(self, rand):
        """Add host at this index to dead list, remove it from alive ones.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual to kill
        """

        index_host,rand = self.getWeightedRandom(
            rand, self.lethality_rate_host
                * self.coefficients_hosts[:,self.LETHALITY]
            )

        self.dead_hosts.append(self.hosts[index_host])
        self.hosts[index_host].die()

    def killVector(self, rand):
        """Add host at this index to dead list, remove it from alive ones.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual to kill
        """

        index_vector,rand = self.getWeightedRandom(
            rand, self.lethality_rate_vector
                * self.coefficients_vectors[:,self.LETHALITY]
            )

        self.dead_vectors.append(self.vectors[index_vector])
        self.vectors[index_vector].die()

    def dieHost(self, rand):
        """Remove host at this index from alive lists.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual to kill
        """

        index_host,rand = self.getWeightedRandom(
            rand, np.ones( self.coefficients_hosts.shape[0] )
            )

        self.hosts[index_host].die()

    def dieVector(self, rand):
        """Remove vector at this index from alive lists.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual to kill
        """

        index_vector,rand = self.getWeightedRandom(
            rand, np.ones( self.coefficients_vectors.shape[0] )
            )

        self.vectors[index_vector].die()

    def birthHost(self, rand):
        """Add host at this index to population, remove it from alive ones.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual to make parent
        """

        index_host,rand = self.getWeightedRandom(
            rand,self.coefficients_hosts[:,self.NATALITY]
            )

        self.hosts[index_host].birth(rand)

    def birthVector(self, rand):
        """Add host at this index to population, remove it from alive ones.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual to make parent
        """

        index_vector,rand = self.getWeightedRandom(
            rand,self.coefficients_vectors[:,self.NATALITY]
            )

        self.vectors[index_vector].birth(rand)

    def mutateHost(self, rand):
        """Mutate a single, random locus in a random pathogen in the given host.

        Creates a new genotype from a de novo mutation event in the host given.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual in which to choose a pathogen to mutate
        """

        index_host,rand = self.getWeightedRandom(
            rand,self.coefficients_hosts[:,self.MUTATION]
            )
        host = self.hosts[index_host]

        host.mutate(rand)

    def mutateVector(self, rand):
        """Mutate a single, random locus in a random pathogen in given vector.

        Creates a new genotype from a de novo mutation event in the vector
        given.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual in which to choose a pathogen to mutate
        """

        index_vector,rand = self.getWeightedRandom(
            rand,self.coefficients_vectors[:,self.MUTATION]
            )
        vector = self.vectors[index_vector]

        vector.mutate(rand)

    def recombineHost(self, rand):
        """Recombine 2 random pathogen genomes at random locus in given host.

        Creates a new genotype from two random possible pathogens in the host
        given.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual in which to choose pathogens to recombine
        """

        index_host,rand = self.getWeightedRandom(
            rand,self.coefficients_hosts[:,self.MUTATION]
            )
        host = self.hosts[index_host]

        host.recombine(rand)

    def recombineVector(self, rand):
        """Recombine 2 random pathogen genomes at random locus in given vector.

        Creates a new genotype from two random possible pathogens in the vector
        given.

        Arguments:
        rand -- uniform random number from 0 to 1 to use when choosing
            individual in which to choose pathogens to recombine
        """

        index_vector,rand = self.getWeightedRandom(
            rand,self.coefficients_vectors[:,self.MUTATION]
            )
        vector = self.vectors[index_vector]

        vector.recombine(rand)

    def updateHostCoefficients(self):
        """Updates event coefficient values in population's hosts."""
        self.coefficients_hosts = np.zeros( self.coefficients_hosts.shape )
        self.coefficients_hosts[ 1:, self.RECEIVE_CONTACT ] = 1
        self.coefficients_hosts[ 1:, self.RECEIVE_POPULATION_CONTACT ] = 1
        self.coefficients_hosts[ 1:, self.NATALITY ] = 1
        self.coefficients_hosts[ 1:, self.MIGRATION ] = 1

        for h in self.hosts:
            genomes = h.pathogens.keys()
            h.pathogens = {}
            h.sum_fitness = 0
            for g in genomes:
                h.acquirePathogen(g)

    def updateVectorCoefficients(self):
        """Updates event coefficient values in population's vectors."""
        self.coefficients_vectors = np.zeros( self.coefficients_vectors.shape )
        self.coefficients_vectors[ 1:, self.RECEIVE_CONTACT ] = 1
        self.coefficients_vectors[ 1:, self.RECEIVE_POPULATION_CONTACT ] = 1
        self.coefficients_vectors[ 1:, self.NATALITY ] = 1
        self.coefficients_vectors[ 1:, self.MIGRATION ] = 1

        for v in self.vectors:
            genomes = v.pathogens.keys()
            v.pathogens = {}
            v.sum_fitness = 0
            for g in genomes:
                v.acquirePathogen(g)

    def healthyCoefficientRow(self):
        """Returns coefficient values corresponding to a healthy host/vector."""

        v = np.zeros((1,self.NUM_COEFFICIENTS))
        v[ 0, self.RECEIVE_CONTACT ] = 1
        v[ 0, self.RECEIVE_POPULATION_CONTACT ] = 1
        v[ 0, self.NATALITY ] = 1
        v[ 0, self.MIGRATION ] = 1

        return v

    def getWeightedRandom(self, rand, r):
        """Returns index of element chosen from weights and given random number.

        Since sampling from coefficient arrays which contain a dummy first row,
        index is decreased by 1.

        Arguments:
        rand -- 0-1 random number (number)
        r -- array with weights (numpy vector)

        Returns:
        new 0-1 random number (number)
        """

        r_tot = np.sum( r )
        u = rand * r_tot # random uniform number between 0 and total rate
        r_cum = 0
        for i,e in enumerate(r): # for every possible event,
            r_cum += e # add this event's rate to cumulative rate
            if u < r_cum: # if random number is under cumulative rate
                return i-1, ( ( u - r_cum + e ) / e )




"""Contains class Host."""

import numpy as np
import copy as cp

class Host(object):
    """Class defines main entities to be infected by pathogens in model.

    Methods:
    copyState -- returns a slimmed-down version of the current host state
    acquirePathogen -- adds given genome to this host's pathogens
    infectHost -- infects given host with a sample of this host's pathogens
    infectVector -- infects given vector with a sample of this host's pathogens
    recover -- removes all infections
    die -- kills this host
    birth -- add a new host to population based on this host
    applyTreatment -- removes all infections with genotypes susceptible to given
        treatment
    mutate -- mutate a single, random locus in a random pathogen
    recombine -- recombine two random pathogen genomes at random locus
    getWeightedRandomGenome -- returns index of element chosen from weights and
        given random number
    """

    def __init__(self, population, id, slim=False):
        """Create a new Host.

        Arguments:
        population -- the population this host belongs to (Population)
        id -- unique identifier for this host within population (String)
        slim -- whether to create a slimmed-down representation of the
            population for data storage (only ID, host and vector lists)
            (Boolean, default False)
        """
        super(Host, self).__init__()
        self.id = id

        if not slim:
                # if not slimmed down for data storage, save other attributes
            self.pathogens = {} # Dictionary with all current infections in this
                # host, with keys=genome strings, values=fitness numbers
            self.protection_sequences = [] # A list of strings this host is
                # immune to. If a pathogen's genome contains one of these
                # values, it cannot infect this host.
            self.population = population
            self.sum_fitness = 0
                # sum of all pathogen fitnesses within this host
            self.coefficient_index = population.coefficients_hosts.shape[0]
                # index in population's coefficient array, not same as id

            population.coefficients_hosts = np.vstack( (
                population.coefficients_hosts,
                population.healthyCoefficientRow()
                ) ) # adds a row to coefficient array

    def copyState(self):
        """Returns a slimmed-down representation of the current host state.

        Returns:
        Host object with current pathogens and protection_sequences.
        """

        copy = Host(None, self.id, slim=True)
        copy.pathogens = self.pathogens.copy()
        copy.protection_sequences = self.protection_sequences.copy()

        return copy

    def acquirePathogen(self, genome):
        """Adds given genome to this host's pathogens.

        Modifies event coefficient matrix accordingly.

        Arguments:
        genome -- the genome to be added (String)
        """
        self.pathogens[genome] = self.population.fitnessHost(genome)
        old_sum_fitness = self.sum_fitness
        self.sum_fitness += self.pathogens[genome]
        sum_fitness_denom = self.sum_fitness if self.sum_fitness > 0 else 1
        self.population.coefficients_hosts[self.coefficient_index,:] = (
            self.population.coefficients_hosts[
                self.coefficient_index,:
                ]
            * old_sum_fitness / sum_fitness_denom ) + ( np.array([
                    # positions dependent on class constants
                0,
                self.population.contactHost(genome),
                self.population.receiveContactHost(genome),
                self.population.lethalityHost(genome),
                self.population.natalityHost(genome),
                self.population.recoveryHost(genome),
                self.population.migrationHost(genome),
                self.population.populationContactHost(genome),
                self.population.receivePopulationContactHost(genome),
                self.population.mutationHost(genome),
                self.population.recombinationHost(genome)
            ]) * self.pathogens[genome] / sum_fitness_denom )

        self.population.coefficients_hosts[
            self.coefficient_index,self.population.RECOMBINATION
            ] = self.population.coefficients_hosts[
                self.coefficient_index,self.population.RECOMBINATION
                ] * ( len(self.pathogens) > 1 )

        self.population.coefficients_hosts[
            self.coefficient_index,self.population.INFECTED
            ] = 1

        if self not in self.population.infected_hosts:
            self.population.infected_hosts.append(self)
            self.population.healthy_hosts.remove(self)

    def infectHost(self, host):
        """Infect given host with a sample of this host's pathogens.

        Each pathogen in the infector is sampled as present or absent in the
        inoculum by drawing from a Poisson distribution with a mean equal to the
        mean inoculum size of the organism being infected weighted by each
        genome's fitness as a fraction of the total in the infector as the
        probability of each trial (minimum 1 pathogen transfered). Each pathogen
        present in the inoculum will be added to the infected organism, if it
        does not have protection from the pathogen's genome. Fitnesses are
        computed for the pathogens' genomes in the infected organism, and the
        organism is included in the poplation's infected list if appropriate.

        Arguments:
        vector -- the vector to be infected (Vector)

        Returns:
        whether or not the model has changed state (Boolean)
        """

        changed = False

        genomes = list( self.pathogens.keys() )
        fitness_weights = [
            self.pathogens[g] / self.sum_fitness for g in genomes
            ]

        genomes_inoculated = np.unique( np.random.choice(
            genomes, p=fitness_weights,
            size=max(
                np.random.poisson( self.population.mean_inoculum_host ), 1
                )
            ) )
        for genome in genomes_inoculated:
            if genome not in host.pathogens.keys() and not any(
                    [ p in genome for p in host.protection_sequences ]
                    ):
                host.acquirePathogen(genome)
                changed = True

        return changed

    def infectVector(self, vector):
        """Infect given host with a sample of this host's pathogens.

        Each pathogen in the infector is sampled as present or absent in the
        inoculum by drawing from a Poisson distribution with a mean equal to the
        mean inoculum size of the organism being infected weighted by each
        genome's fitness as a fraction of the total in the infector as the
        probability of each trial (minimum 1 pathogen transfered). Each pathogen
        present in the inoculum will be added to the infected organism, if it
        does not have protection from the pathogen's genome. Fitnesses are
        computed for the pathogens' genomes in the infected organism, and the
        organism is included in the poplation's infected list if appropriate.

        Arguments:
        vector -- the vector to be infected (Vector)

        Returns:
        whether or not the model has changed state (Boolean)
        """

        changed = False

        genomes = list( self.pathogens.keys() )
        fitness_weights = [
            self.pathogens[g] / self.sum_fitness for g in genomes
            ]

        genomes_inoculated = np.unique( np.random.choice(
            genomes, p=fitness_weights,
            size=max(
                np.random.poisson( self.population.mean_inoculum_vector ), 1
                )
            ) )
        for genome in genomes_inoculated:
            if genome not in vector.pathogens.keys() and not any(
                    [ p in genome for p in vector.protection_sequences ]
                    ):
                vector.acquirePathogen(genome)
                changed = True

        return changed

    def recover(self):
        """Remove all infections from this host.

        If model is protecting upon recovery, add protecion sequence as defined
        by the indexes in the corresponding model parameter. Remove from
        population infected list and add to healthy list.
        """

        if self in self.population.infected_hosts:
            if self.population.protection_upon_recovery_host:
                for genome in self.pathogens:
                    seq = genome[
                        self.population.protection_upon_recovery_host[0]
                        :self.population.protection_upon_recovery_host[1]
                        ]
                    if seq not in self.protection_sequences:
                        self.protection_sequences.append(seq)

            self.pathogens = {}
            self.sum_fitness = 0
            self.population.coefficients_hosts[
                self.coefficient_index,:
                ] = self.population.healthyCoefficientRow()

            self.population.infected_hosts.remove(self)
            self.population.healthy_hosts.append(self)

    def die(self):
        """Add host to population's dead list, remove it from alive ones."""

        if self in self.population.infected_hosts:
            self.population.infected_hosts.remove(self)
        else:
            self.population.healthy_hosts.remove(self)

        for h in self.population.hosts[self.coefficient_index:]:
            h.coefficient_index -= 1

        self.population.coefficients_hosts = np.delete(
            self.population.coefficients_hosts, self.coefficient_index, 0
            )
        self.population.hosts.remove(self)

    def birth(self, rand):
        """Add a new host to population based on this host."""

        host_list = self.population.addHosts(1)
        host = host_list[0]

        if self.population.vertical_transmission_host > rand:
            self.infectHost(host)

        if self.population.inherit_protection_host > np.random.random():
            host.protection_sequences = self.protection_sequences.copy()

    def applyTreatment(self, resistance_seqs):
        """Remove all infections with genotypes susceptible to given treatment.

        Pathogens are removed if they are missing at least one of the sequences
        in resistance_seqs from their genome. Removes this organism from
        population infected list and adds to healthy list if appropriate.

        Arguments:
        resistance_seqs -- contains sequences required for treatment resistance
            (list of Strings)
        """

        genomes_remaining = []
        for genome in self.pathogens:
            for seq in resistance_seqs:
                if seq in genome:
                    genomes_remaining.append(genome)
                    break

        if len(genomes_remaining) == 0:
            self.recover()
        else:
            self.pathogens = {}
            self.sum_fitness = 0
            self.population.coefficients_hosts[
                self.coefficient_index,:
                ] = self.population.healthyCoefficientRow()
            for genome in genomes_remaining:
                self.acquirePathogen(genome)

    def mutate(self,rand):
        """Mutate a single, random locus in a random pathogen.

        Creates a new genotype from a de novo mutation event.
        """

        genomes = list( self.pathogens.keys() )
        weights = [
            self.population.mutationHost(g)
            * self.population.fitnessHost(g) for g in genomes
            ]
        index_genome,rand = self.getWeightedRandomGenome( rand,weights )

        old_genome = genomes[index_genome]
        mut_index = int( rand * self.population.num_loci )
        if old_genome[mut_index] != self.population.CHROMOSOME_SEPARATOR:
            new_genome = old_genome[0:mut_index] + np.random.choice(
                list(self.population.possible_alleles[mut_index])
                ) + old_genome[mut_index+1:]
            if new_genome not in self.pathogens:
                self.acquirePathogen(new_genome)

            if new_genome not in self.population.model.global_trackers['genomes_seen']:
                self.population.model.global_trackers['genomes_seen'].append(new_genome)

    def recombine(self,rand):
        """Recombine two random pathogen genomes at random locus.

        Creates a new genotype from two random possible pathogens.
        """

        genomes = list( self.pathogens.keys() )
        weights = [
            self.population.recombinationHost(g)
            * self.population.fitnessHost(g) for g in genomes
            ]
        index_genome,rand = self.getWeightedRandomGenome( rand,weights )
        index_other_genome,rand = self.getWeightedRandomGenome( rand,weights )

        if index_genome != index_other_genome:
            num_evts = np.random.poisson( self.population.num_crossover_host )
            loci = np.random.randint( 0, self.population.num_loci, num_evts )

            children = [ genomes[index_genome], genomes[index_other_genome] ]

            for l in loci:
                children[0] = children[0][0:l] + children[1][l:]
                children[1] = children[1][0:l] + children[0][l:]

            children = [
                genome.split(self.population.CHROMOSOME_SEPARATOR)
                for genome in children
                ]
            parent = np.random.randint( 0, 2, len( children[0] ) )

            children = [
                self.population.CHROMOSOME_SEPARATOR.join([
                    children[ parent[i] ][i]
                    for i in range( len( children[0] ) )
                    ]),
                self.population.CHROMOSOME_SEPARATOR.join([
                    children[ not parent[i] ][i]
                    for i in range( len( children[1] ) )
                    ])
                ]

            for new_genome in children:
                if new_genome not in self.pathogens:
                    self.acquirePathogen(new_genome)

                if new_genome not in self.population.model.global_trackers['genomes_seen']:
                    self.population.model.global_trackers['genomes_seen'].append(new_genome)

    def getWeightedRandomGenome(self, rand, r):
        """Returns index of element chosen from weights and given random number.

        Arguments:
        rand -- 0-1 random number (number)
        r -- array with weights (numpy vector)

        Returns:
        new 0-1 random number (number)
        """

        r_tot = np.sum( r )
        u = rand * r_tot # random uniform number between 0 and total rate
        r_cum = 0
        for i,e in enumerate(r): # for every possible event,
            r_cum += e # add this event's rate to cumulative rate
            if u < r_cum: # if random number is under cumulative rate
                return i, ( ( u - r_cum + e ) / e )
